<?php

namespace Slack\Listeners;

use Illuminate\Contracts\Queue\ShouldQueue;
use Illuminate\Support\Facades\Log;

use Slack\Jobs\DeleteChat;
use Slack\Interactive\BlockActions;

/**
 * This class handles BlockActions events.
 * It's expected that the local application would implement this class completely, rather than using this
 * module's implementation.
 *
 * @note Since block actions contain a trigger_id, we shouldnt queue this as the trigger_id may have expired by the time
 *       the queue runs the job. (trigger_id's only last 3 seconds)
 */
class BlockActionListener //implements ShouldQueue
{
	protected const LOGKEY = 'LBA';

	public $queue = 'slack';

	/**
	 * Handle the event.
	 *
	 * @param BlockActions $event
	 * @return void
	 * @throws \Exception
	 */
	public function handle(BlockActions $event): void
	{
		// Do some magic with event data
		Log::info(sprintf('%s:Block Action for Callback [%s] User [%s] in [%s]',self::LOGKEY,$event->callback_id,$event->user_id,$event->team_id),['m'=>__METHOD__]);

		switch ($event->callback_id) {
			// Messages that generate a block action dont have a callback ID
			case NULL:
				Log::debug(sprintf('%s:Callback NULL [%s] (%s)',self::LOGKEY,$event->isMessage(),$event->action_id),['m'=>__METHOD__]);

				$this->messageEvent($event);
				break;

			default:
				Log::notice(sprintf('%s:Unhandled CALLBACK [%s]',self::LOGKEY,$event->callback_id),['m'=>__METHOD__]);
		}
	}

	/**
	 * Events from messages
	 *
	 * @param BlockActions $event
	 * @throws \Exception
	 */
	protected function messageEvent(BlockActions $event): void
	{
		Log::info(sprintf('%s:[%s] has [%d] actions',self::LOGKEY,$event->callback_id,count($event->actions)),['m'=>__METHOD__]);

		foreach ($event->actions as $id => $action) {
			$event->index = $id;

			Log::debug(sprintf('%s:Action [%s]',self::LOGKEY,$event->action_id),['m'=>__METHOD__]);
			switch ($event->action_id) {
				case 'self_destruct':
					// Queue the delete of the message
					dispatch((new DeleteChat($event->user(),$event->message_ts))->onQueue('slack'));

					// @todo If this message is on integrations messages channel, which is not the user_id() - need to use the user's integration direct channel ID
					break;

				default:
					Log::notice(sprintf('%s:Unhandled ACTION [%s]',self::LOGKEY,$event->action_id),['m'=>__METHOD__]);
			}
		}
	}
}